#include "tgxf-transmit-pull.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/stat.h>
#ifdef _WIN32
# include <io.h>
# include <fcntl.h>
#endif
#include <math.h>
//#include <gd.h>
//#include <png.h>


// The following CRC code comes from RFC1952, Section 8
//   http://tools.ietf.org/html/rfc1952#section-8

/* Table of CRCs of all 8-bit messages. */
static unsigned long crc_table[256];

/* Flag: has the table been computed? Initially false. */
static int crc_table_computed = 0;

/* Make the table for a fast CRC. */
static void make_crc_table(void)
{
  unsigned long c;
  int n, k;
  for (n = 0; n < 256; n++) {
    c = (unsigned long) n;
    for (k = 0; k < 8; k++) {
      if (c & 1) {
        c = 0xedb88320L ^ (c >> 1);
      } else {
        c = c >> 1;
      }
    }
    crc_table[n] = c;
  }
  crc_table_computed = 1;
}


/*
    Update a running crc with the bytes buf[0..len-1] and return
  the updated crc. The crc should be initialized to zero. Pre- and
  post-conditioning (one's complement) is performed within this
  function so it shouldn't be done by the caller. Usage example:

  unsigned long crc = 0L;

  while (read_buffer(buffer, length) != EOF) {
    crc = update_crc(crc, buffer, length);
  }
  if (crc != original_crc) error();
*/
static unsigned long update_crc(unsigned long crc,
  unsigned char *buf, int len)
{
  unsigned long c = crc ^ 0xffffffffL;
  int n;

  if (!crc_table_computed)
    make_crc_table();
  for (n = 0; n < len; n++) {
    c = crc_table[(c ^ buf[n]) & 0xff] ^ (c >> 8);
  }
  return c ^ 0xffffffffL;
}

/* Return the CRC of the bytes buf[0..len-1]. */
static unsigned long crc(unsigned char *buf, int len)
{
  return update_crc(0L, buf, len);
}




/*

                                   _JNJ`
                               .JNMH`
                              JMMF`       `;.
                            .NMM)          `MN.
                            MMM)            (MML
                           (MMM`             MMML
                     M  I  D  N  I  G  H  T    C  o  D  E
                           (NMMF             MHNH
                            NMML            .MMM
                             NMML          .NMH
                              4MMNL       .#F
                               `4HNNL_   `
                                   `"""`

                          Copyright (C) 2004-2014
       "Ian (Larry) Latter" <ian dot latter at midnightcode dot org>

    Midnight Code is a registered trademark of Ian Latter.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, and mirrored at the Midnight Code web
    site; as at version 2 of the License only.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    (version 2) along with this program; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA, or see http://midnightcode.org/gplv2.txt

*/

#define TGXF_DEBUG 0

#define TGXF_CONTROL_TYPE_START 1
#define TGXF_CONTROL_TYPE_STOP 2
#define TGXF_CONTROL_TYPE_STATUS 3
#define TGXF_CONTROL_SUBTYPE_START_FILENAME 1
#define TGXF_CONTROL_SUBTYPE_START_FILESIZE 2
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_VERSION 3
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_FPS 4
#define TGXF_CONTROL_SUBTYPE_START_QRCODE_BYTES 5
#define TGXF_CONTROL_SUBTYPE_STOP_PAUSE 1
#define TGXF_CONTROL_SUBTYPE_STOP_COMPLETE 2
#define TGXF_CONTROL_SUBTYPE_STOP_CANCEL 3
#define TGXF_CONTROL_SUBTYPE_STATUS_SINCE 1
#define TGXF_DISPLAY_ASCII_SIMPLE 1
#define TGXF_DISPLAY_ASCII_SIMPLE_SQUARE 2
#define TGXF_DISPLAY_ASCII_COMPRESSED 3
#define TGXF_DISPLAY_ANSI_SIMPLE 4
#define TGXF_DISPLAY_ANSI_SIMPLE_SQUARE 5

// TGXf Session Parameters (TGXf State Machine)
static unsigned int tgxf_session_total_frames = 0;
static unsigned int tgxf_session_frame_count = 0;
static unsigned long tgxf_session_total_crc32 = 0L;


static unsigned int tgxf_session_debug_image_counter = 0;

// TGXf Session Parameters (TGXf Global Options)
static unsigned int tgxf_session_read_bytes = 0;
static const char * tgxf_session_filepath;
static unsigned int tgxf_session_file_size = 0;
static unsigned int tgxf_session_qrcode_version = 0;
static unsigned int tgxf_session_qrcode_fps = 0;
static unsigned int tgxf_session_qrcode_bytes = 0;

static unsigned char * tgxf_packet;

static char * tgxf_session_data_to_send = 0;

unsigned int tgxf_get_num_frames() {
    return tgxf_session_total_frames;
}

static char * get_basename_for_file(char *filepath) {
  char *basename;

  basename = filepath;
  while(*filepath) {
    if(*filepath++ == '/')
      basename = filepath;
  }

  return basename;
}


static unsigned long get_crc_for_data(unsigned char *data, int len) {
    int data_buffer_size = 1024;
    int read = 0;
    unsigned long crc;

    crc = 0L;
    
    while (read < len) {
        int numToRead = ((len-read)< data_buffer_size) ? (len-read) : data_buffer_size;
        crc = update_crc(crc, data, numToRead);
        
        data += numToRead;
        read += numToRead;
    }
    
    return crc;
}


static unsigned int get_control_bit(unsigned char *data) {
  unsigned char control_byte;
  unsigned int control_bit;

  control_byte = *data;
  control_bit = control_byte & 1;

  return control_bit;
}


static unsigned int get_control_type(unsigned char *data) {
  unsigned char control_byte;
  unsigned int control_type;

  control_byte = *data;
  control_type = (control_byte & 14) >> 1;

  return control_type;
}


static unsigned int pack_control_payload_16bit_number(unsigned char *payload, void *data) {
  unsigned int value;
  unsigned int *value_p;

  if(!payload || !data)
    return 0;

  value_p = (unsigned int *)data;
  value = *value_p;
  *payload = (value >> 8) & 0xFF;
  payload++;
  *payload = (value >> 0) & 0xFF;

  return value;
}


static unsigned long pack_control_payload_32bit_number(unsigned char *payload, void *data) {
  unsigned long value;
  unsigned long *value_p;

  if(!payload || !data)
    return 0;

  value_p = (unsigned long *)data;
  value = *value_p;
  *payload = (value >> 24) & 0xFF;
  payload++;
  *payload = (value >> 16) & 0xFF;
  payload++;
  *payload = (value >> 8) & 0xFF;
  payload++;
  *payload = (value >> 0) & 0xFF;

  return value;
}


static unsigned int pack_control_payload_string(unsigned char *payload, void *data, unsigned int data_len) {

  if(!payload || !data || !data_len)
    return 0;

  memset(payload, 0, 9);
  if(data_len > 9)
    data_len = 9;
  memcpy(payload, data, data_len);

  return data_len;
}


static unsigned int pack_data_payload_bin(unsigned char *payload, void *data, unsigned int data_len) {

  if(!payload || !data || !data_len)
    return 0;

  memcpy(payload, data, data_len);

  return data_len;
}


static unsigned char * build_control_frame(unsigned char * packet, unsigned int control_type, unsigned int control_subtype, void * data, unsigned int length) {
  unsigned char control_byte;
  unsigned int control_bit;
  unsigned char * payload;

  // Zero Control Byte 
  control_byte = 0;

  // Control Frame, Control Bit = 1 (bit 0); 
  control_bit = 1;
  control_byte = control_byte | control_bit;

  // Control Byte has Control Type (bits 3,2,1) 
  if(control_type < 0 || control_type >= 7)
    control_type = 0;
  control_type = control_type << 1;
  control_byte = control_byte | control_type;

  // Control Byte has Sub-Control Type (bits 7,6,5,4) 
  if(control_subtype < 0 || control_subtype >= 15)
    control_subtype = 0;
  control_subtype = control_subtype << 4;
  control_byte = control_byte | control_subtype;

  payload = packet + 1;

  switch((control_type & 14) >> 1) {
    case TGXF_CONTROL_TYPE_START:
      switch((control_subtype & 240) >> 4) {
       case TGXF_CONTROL_SUBTYPE_START_FILENAME:                    // FILENAME 
          pack_control_payload_string(payload, data, length);
          break;
       case TGXF_CONTROL_SUBTYPE_START_FILESIZE:                    // FILESIZE 
          pack_control_payload_16bit_number(payload, data);
          break;
       case TGXF_CONTROL_SUBTYPE_START_QRCODE_VERSION:              // QRCODE_VERSION 
          pack_control_payload_16bit_number(payload, data);
          break;
       case TGXF_CONTROL_SUBTYPE_START_QRCODE_FPS:                  // QRCODE_FPS 
          pack_control_payload_16bit_number(payload, data);
          break;
       case TGXF_CONTROL_SUBTYPE_START_QRCODE_BYTES:                // QRCODE_BYTES 
          pack_control_payload_16bit_number(payload, data);
          break;
      }
      break;

    case TGXF_CONTROL_TYPE_STOP:
      switch((control_subtype & 240) >> 4) {
       case TGXF_CONTROL_SUBTYPE_STOP_PAUSE:                        // PAUSE
          break;
       case TGXF_CONTROL_SUBTYPE_STOP_COMPLETE:                     // COMPLETE
          pack_control_payload_32bit_number(payload, data);
          break;
       case TGXF_CONTROL_SUBTYPE_STOP_CANCEL:                       // CANCEL
          pack_control_payload_string(payload, data, length);
          break;
      }
      break;

    case TGXF_CONTROL_TYPE_STATUS:
      switch((control_subtype & 240) >> 4) {
       case TGXF_CONTROL_SUBTYPE_STATUS_SINCE:                      // SINCE
          pack_control_payload_32bit_number(payload, data);
          break;
      }
      break;

    default:
      break;
  }

  // Frame consists of Control Byte and Control Payload 
  *packet = (control_byte) & 0xFF;

  return packet;
}


static unsigned char * build_data_frame(unsigned char * packet, unsigned int counter, void * data, unsigned int payload_size) {
  unsigned char control_byte;
  unsigned int control_bit;
  unsigned char * payload;

  // Zero Control Byte 
  control_byte = 0;

  // Data Frame, Control Bit = 0 (bit 0); 
  control_bit = 0;
  control_byte = control_byte | control_bit;

  // Control Byte has incremented counter (bits 4,3,2,1) 
  if(counter < 0 || counter > 15)
    counter = 0;
  counter = counter << 1;
  control_byte = control_byte | counter;

  // Frame consists of Control Byte and Data Payload 
  payload = packet + 1;
  *packet = (control_byte) & 0xFF;
  if(memcpy(payload, data, payload_size) != payload) {
    // fprintf(stderr, "ERROR(TGXf): failed to build frame\n");
    // What else can we do here?
    return packet;
  }

  return packet;
}

// JAMES - version which just fills in a bitmap, no image manipulation
unsigned char * get_frame(unsigned char * data, int length, int qr_ver, int qr_ecc, QRcode ** code) {
    // Generate QRCode as an array of 1 and 0 values, from 8bit data
    *code = QRcode_encodeData(length, data, qr_ver, qr_ecc);
    
    tgxf_session_debug_image_counter++;

    return data;
}


unsigned int pull_frame(unsigned int frame_number,
  unsigned char * tgxf_packet, unsigned int read_bytes,
  int mode_ecc_level, int mode_pixel_size, int mode_margin_size, QRcode ** code) {

  char * basename;
  unsigned int last_frame;
  unsigned int data_frame_number;
  unsigned int start_control_frames;
  int read_delta;
  FILE *fp;

  start_control_frames = 4;
  last_frame = tgxf_session_total_frames + start_control_frames + 1;
  fprintf(stderr, "TEST(TGXf): frame_no[%d], last_frame[%d], tgxf_sess..count=[%d]\n",
    frame_number, last_frame, tgxf_session_total_frames);

  switch(frame_number) {
    case 1:
      // TGXf CONTROL -> START -> FILENAME
      basename = get_basename_for_file(tgxf_session_filepath);
      memset(tgxf_packet, 0, tgxf_session_qrcode_bytes);
      build_control_frame(
        tgxf_packet, 
        TGXF_CONTROL_TYPE_START,
        TGXF_CONTROL_SUBTYPE_START_FILENAME,
        basename, strlen(basename));
      get_frame(tgxf_packet, tgxf_session_qrcode_bytes, tgxf_session_qrcode_version,
        mode_ecc_level, code);
      break;

    case 2:
      // TGXf CONTROL -> START -> FILESIZE
      memset(tgxf_packet, 0, tgxf_session_qrcode_bytes);
      build_control_frame(
        tgxf_packet, 
        TGXF_CONTROL_TYPE_START,
        TGXF_CONTROL_SUBTYPE_START_FILESIZE,
        &tgxf_session_file_size, 3);
      get_frame(tgxf_packet, tgxf_session_qrcode_bytes, tgxf_session_qrcode_version,
        mode_ecc_level, code);
      break;

    case 3:
      // TGXf CONTROL -> START -> QRCODE_BYTES
      memset(tgxf_packet, 0, tgxf_session_qrcode_bytes);
      build_control_frame(
        tgxf_packet, 
        TGXF_CONTROL_TYPE_START,
        TGXF_CONTROL_SUBTYPE_START_QRCODE_BYTES,
        &read_bytes, 3);
      get_frame(tgxf_packet, tgxf_session_qrcode_bytes, tgxf_session_qrcode_version,
        mode_ecc_level, code);
      break;

    case 4:
      // TGXf CONTROL -> START -> QRCODE_FPS
      memset(tgxf_packet, 0, tgxf_session_qrcode_bytes);
      build_control_frame(
        tgxf_packet, 
        TGXF_CONTROL_TYPE_START,
        TGXF_CONTROL_SUBTYPE_START_QRCODE_FPS,
        &tgxf_session_qrcode_fps, 3);
      get_frame(tgxf_packet, tgxf_session_qrcode_bytes, tgxf_session_qrcode_version,
        mode_ecc_level, code);
      break;

    default:
      if(frame_number < last_frame) {
        // TGXf DATA (one data frame per loop iteration)
        data_frame_number = frame_number - start_control_frames - 1;

        int data_start_index = (read_bytes * data_frame_number);
          
        // last block an odd size?
        read_delta = tgxf_session_file_size - (data_frame_number * read_bytes);
        if(read_delta > 0 &&
          read_delta < read_bytes &&
          data_frame_number == (tgxf_session_total_frames - 1)) {
            read_bytes = read_delta;
        }

        memset(tgxf_packet, 0, tgxf_session_qrcode_bytes);
        build_data_frame(tgxf_packet, tgxf_session_frame_count, tgxf_session_data_to_send + data_start_index, read_bytes);
        get_frame(tgxf_packet, tgxf_session_qrcode_bytes, tgxf_session_qrcode_version,
        mode_ecc_level, code);

        tgxf_session_frame_count++;
        if(tgxf_session_frame_count > 15)
          tgxf_session_frame_count = 0;
      } else {
        // TGXf CONTROL -> STOP -> COMPLETE
        memset(tgxf_packet, 0, tgxf_session_qrcode_bytes);
        build_control_frame(
          tgxf_packet, 
          TGXF_CONTROL_TYPE_STOP,
          TGXF_CONTROL_SUBTYPE_STOP_COMPLETE,
          &tgxf_session_total_crc32, 5);
        get_frame(tgxf_packet, tgxf_session_qrcode_bytes, tgxf_session_qrcode_version,
          mode_ecc_level, code);
      }
      break;
  }

  return 0;
}

unsigned int tgxf_pull_frame(unsigned int frame_number, QRcode ** code) {
    
    return pull_frame(frame_number, tgxf_packet, tgxf_session_read_bytes,
               QR_ECLEVEL_M, 3, 1, code);
}

int tgxf_init_session(char * bytesIn, int length, unsigned int qrcodeVersion, const char *filePath, int fps)
{
    unsigned int frame_rounding_correction;
    unsigned int frame_bytes_version[64];
    
    if(qrcodeVersion != 1 && qrcodeVersion != 2 && qrcodeVersion != 8 && qrcodeVersion != 15) {
        fprintf(stderr, "WARNING(TGXf): qr code version %d not supported. Using default\n", qrcodeVersion);
        qrcodeVersion = 8;
    }
    
    tgxf_session_file_size = length;
    tgxf_session_total_crc32 = get_crc_for_data(bytesIn, length);
    tgxf_session_data_to_send = bytesIn;
    tgxf_session_qrcode_fps = fps;
    
    tgxf_session_qrcode_version = qrcodeVersion;
    tgxf_session_filepath = filePath;
    
    frame_rounding_correction = 4;
    frame_bytes_version[1] = 14 - frame_rounding_correction;
    frame_bytes_version[2] = 26 - frame_rounding_correction;
    frame_bytes_version[8] = 152 - frame_rounding_correction;
    frame_bytes_version[15] = 412 - frame_rounding_correction;
    tgxf_session_qrcode_bytes = frame_bytes_version[tgxf_session_qrcode_version];
    
    tgxf_session_read_bytes = tgxf_session_qrcode_bytes - 1;  // Subtract the Control Byte
    
    
    // Initialisation
    tgxf_session_frame_count = 0;
    if((tgxf_packet = (unsigned char *)malloc(tgxf_session_qrcode_bytes)) == NULL) {
        fprintf(stderr, "ERROR(TGXf): unable to allocate frame buffer\n");
        return -1;
    }
    tgxf_session_total_frames = tgxf_session_file_size / tgxf_session_read_bytes;
    if(tgxf_session_total_frames * tgxf_session_read_bytes < tgxf_session_file_size) {
        tgxf_session_total_frames++;
    }
    
    return 0;
}



